from rq_interfaces.msg import Projection, ProjectionGeometry, CheckedScanPose
from rq_interfaces.srv import AquireProjection, ReachabilityCheck, ReachabilityMapCheck
from sensor_msgs.msg import Image

from rclpy.node import Node
from pathlib import Path
from tf2_ros import TransformBroadcaster

import numpy as np
from geometry_msgs.msg import TransformStamped
import ros2_numpy

from . import PROJECTION_TEMP_FOLDER


class BaseHardwareInterfaceService(Node):
    """
    BaseHardwareInterfaceService is a ROS 2 Node that provides a base interface for hardware interaction services.

    Attributes:
        interface_type (str): The type of interface.
        last_projection (PyProjection): The last received projection.
        measurement_name (str): The name of the measurement.
        projection_tempfolder (Path): The temporary folder for storing projections.
    """

    interface_type: str = "Base"
    last_projection_image: np.ndarray = None
    measurement_name: str = "projection"
    projection_tempfolder: Path = PROJECTION_TEMP_FOLDER

    def __init__(self, nodename: str = "rq_base_hardware_interface_service"):
        """
        Initialize the BaseHardwareInterfaceService node, create services and publishers, and set up a subscription.

        :param node_name: The name of the node. Defaults to 'rq_base_hardware_interface_service'.
        """
        super().__init__(nodename)

        self.aquire_projection_srv = self.create_service(
            AquireProjection, "aquire_projection", self.aquire_projection_callback
        )
        self.reachability_srv = self.create_service(
            ReachabilityCheck, "check_reachability", self.check_reachability_callback
        )
        self.reachability_map_srv = self.create_service(
            ReachabilityMapCheck,
            "check_reachability_map",
            self.check_reachability_map_callback,
        )

        self.projection_publisher = self.create_publisher(
            Image, "current_projection_image", 5
        )
        self.projection_geometry_publisher = self.create_publisher(
            ProjectionGeometry, "current_projection_geometry", 5
        )
        self.reachabilty_publisher = self.create_publisher(
            CheckedScanPose, "checked_scan_pose", 5
        )

        self.projection_subscriper = self.create_subscription(
            Image, "current_projection_image", self.projection_callback, 5
        )
        self.transform_broadcaster = TransformBroadcaster(
            Node("rq_transform_broadcaster")
        )

    def aquire_projection_callback(
        self, request: AquireProjection.Request, response: AquireProjection.Response
    ):
        """
        Handle the aquire_projection service request.

        :param request: The service request for acquiring a projection.
        :param response: The service response to be filled.
        :raises NotImplementedError: This method should be overridden by a subclass.
        """
        raise NotImplementedError

    def check_reachability_callback(
        self, request: ReachabilityCheck.Request, response: ReachabilityCheck.Response
    ):
        """
        Handle the check_reachability service request.

        :param request: The service request for checking reachability.
        :param response: The service response to be filled.
        :raises NotImplementedError: This method should be overridden by a subclass.
        """
        raise NotImplementedError

    def check_reachability_map_callback(
        self,
        request: ReachabilityMapCheck.Request,
        response: ReachabilityMapCheck.Response,
    ):
        """
        Handle the check_reachability_map service request.

        :param request: The service request for checking the reachability map.
        :param response: The service response to be filled.
        :raises NotImplementedError: This method should be overridden by a subclass.
        """
        raise NotImplementedError

    def projection_callback(self, msg: Projection):
        """
        Handle incoming Projection messages.

        :param msg: The received Projection message.
        """
        self.get_logger().info("got projection ...")
        self.last_projection_image = ros2_numpy.numpify(msg)

    def set_source_pose(self, position, quarternion):
        t = TransformStamped()
        t.header.stamp = self.get_clock().now().to_msg()
        t.header.frame_id = "object"
        t.child_frame_id = "rq_focal_spot"

        t.transform.translation.x = position[0]
        t.transform.translation.y = position[1]
        t.transform.translation.z = position[2]

        t.transform.rotation.x = quarternion[0]
        t.transform.rotation.y = quarternion[1]
        t.transform.rotation.z = quarternion[2]
        t.transform.rotation.w = quarternion[3]

        self.transform_broadcaster.sendTransform(t)

    def set_detector_pose(self, position, quarternion):
        t = TransformStamped()
        t.header.stamp = self.get_clock().now().to_msg()
        t.header.frame_id = "object"
        t.child_frame_id = "rq_detector"

        t.transform.translation.x = position[0]
        t.transform.translation.y = position[1]
        t.transform.translation.z = position[2]

        t.transform.rotation.x = quarternion[0]
        t.transform.rotation.y = quarternion[1]
        t.transform.rotation.z = quarternion[2]
        t.transform.rotation.w = quarternion[3]

        self.transform_broadcaster.sendTransform(t)
